#!/bin/bash

APP_VERSION=1.0.0

if [[ -z $FUSION_PRODUCT ]]; then
    FUSION_PRODUCT="fusion"
fi

cd $(dirname $0)

source installer.inc.sh

function show_usage()
{
    cat << EOF
Usage: $0 <app-webserver-port> <fusion-admin-username> <fusion-admin-password> <smtp-server> <smtp-username> <smtp-password> <smtp-sender>

  app-webserver-port     Port that the app webserver runs on
  fusion-admin-username  The username for a Fusion admin
  fusion-admin-password  The password for the specified Fusion admin
  smtp-server            SMTP server the service will use
  smtp-username          Username for the SMTP server
  smtp-password          Password for the SMTP server
  smtp-sender            Sender for the SMTP server (optional, defaults to smtp-username)

EOF
    exit 1
}

function install_failed()
{
    err_exit "Something went wrong installing $app_name"
}

[[ $# -lt 6 ]] && show_usage
civetweb_port=$1
fusion_user=$2
fusion_password=$3
smtp_server=$4
smtp_username=$5
smtp_password=$6
smtp_sender=$7
if [[ -z $smtp_sender ]]; then
    smtp_sender=$smtp_username
fi

check_sudo
deps=(curl jq python3)
check_deps_and_install ${deps[@]}

fusion_properties=/etc/opt/${FUSION_PRODUCT}/${FUSION_PRODUCT}.properties
if [[ ! -f $fusion_properties ]]; then
    err_exit "$fusion_properties does not exist."
fi

fusion_apps_dir=$(get_fusion_apps_dir)
if [[ -z $fusion_apps_dir ]]; then
    fusion_apps_dir=/etc/opt/${FUSION_PRODUCT}/apps
    echo "Note: fusion.apps.path is missing from Fusion properties, setting it to $fusion_apps_dir"
    echo "fusion.apps.path=$fusion_apps_dir" >> $fusion_properties
fi

app_dir=$fusion_apps_dir/$app_name
app_config_dir=$app_dir/conf
app_service_dir=$app_dir/service
app_service_config=$app_service_dir/config.ini
app_database=$app_service_dir/$app_name.sqlite
app_config_template=config.ini.template
app_service=$app_name.service
app_service_template=$app_service.template
app_service_log=$app_service_dir/logs/$app_name.log
civetweb_config_template=civetweb.conf.template
civetweb_service=civetweb_$app_name.service
civetweb_service_template=$civetweb_service.template
civetweb_service_dir=$app_dir/webserver
civetweb_root_dir=$civetweb_service_dir/root
config_template=app.json.template

# Make sure fusion is running
pong=$(curl -ksf "$(get_fusion_url $fusion_user $fusion_password)/service/ping")
if [[ $pong != "pong!" ]]; then 
    err_exit "Could not connect to Fusion, please make sure it is running."
fi

# If the app has already been installed we want to stop it
systemctl stop $app_service &> /dev/null || true
systemctl stop $civetweb_service &> /dev/null || true

# Check that the port is not being used.
port_check=$(lsof -i TCP:$civetweb_port | grep LISTEN)
if [[ $? -eq 0 ]]; then
    err_exit "port $civetweb_port is in use."
fi

# If motion database exists preserve it
if [[ -f $app_database ]]; then 
    echo
    echo "Preserving old database $app_database."
    database_save="/tmp/$(basename $app_database)"
    cp $app_database $database_save
fi

# If config exists back it up
if [[ -f $app_service_config ]]; then 
    echo 
    echo "Backing up the current service configuration."
    echo "Old service configuration is not automatically migrated." 
    echo "Any configuration migration will need to be done manually."
    motion_service_config_backup="/tmp/config_$(date +%s).ini.backup"
    cp $app_service_config $motion_service_config_backup
fi

# From this point on trap all errors
trap "install_failed" ERR

# Copy the app to the apps dir
echo
echo Copying app to: $fusion_apps_dir.
rm -rf $app_dir
mkdir -p $app_dir/conf
cp -r ../service ../webserver $app_dir
[[ -f $database_save ]] && mv $database_save $app_service_dir
[[ -f $motion_service_config_backup ]] && mv $motion_service_config_backup $app_service_dir

# Create standalone uninstaller.
# The 'declare -f' statement prints all function definitions within the current scope
# to stdout. So we filter that output to the functions we need from installer.inc.sh
# and write them directly to the uninstall script.
uninstall_sh=$(readlink -f ..)/uninstall.sh # place it in parent directory
uninstall_func_deps=(err_exit check_sudo get_property_value \
    get_fusion_apps_dir get_fusion_url remove_fusion_user)
echo
echo Creating uninstaller at $(dirname $uninstall_sh).
cat << EOF > $uninstall_sh
#!/bin/bash

show_usage()
{
    cat << EOF_UNINSTALL_SH
Usage: \$0 <fusion-admin-username> <fusion-admin-password>

  fusion-admin-username  The username of a Fusion admin
  fusion-admin-password  The password for the specified Fusion admin

EOF_UNINSTALL_SH
    exit 1
}

$(declare -f ${uninstall_func_deps[@]})

[[ \$# -lt 2 ]] && show_usage
fusion_user=\$1
fusion_pass=\$2
services=($app_service $civetweb_service)

check_sudo

# Remove the services
for service in \${services[@]}; do
    systemctl disable --now \$service
    rm /etc/systemd/system/\$service
done
systemctl daemon-reload
systemctl reset-failed

remove_fusion_user \$fusion_user \$fusion_pass $app_name

fusion_apps_dir=\$(get_fusion_apps_dir)
if [[ -z \$fusion_apps_dir ]]; then
    err_exit "fusion.apps.path is missing from Fusion properties, can't cleanup $app_name"
fi
cd \$fusion_apps_dir
rm -rf \$fusion_apps_dir/$app_name

# Unload app from fusion
systemctl restart $FUSION_PRODUCT
EOF
chmod 755 $uninstall_sh
echo Copying uninstaller to $app_dir.
cp $uninstall_sh $app_dir

# Create the application's fusion user (removing any existing account first)
app_user=$app_name
remove_fusion_user $fusion_user $fusion_password $app_user
echo
echo "Creating Fusion user: $app_user."
app_user_password=$(head /dev/urandom | tr -dc A-Za-z0-9 | head -c 16)
create_fusion_user $fusion_user $fusion_password $app_user $app_user_password

echo 
echo "Installing configuration files."
# Create and install the app config
declare -A app_config_replace
app_config_replace[replace_me_with_port]=$civetweb_port
app_config_replace[replace_me_with_app_version]=$APP_VERSION
app_config=$(create_file_from_template "$config_template" "app_config_replace")
mv $app_config $app_config_dir

# Create and install configuration file for the motion alerter service
#TODO(elittley) remove motion from the name to make it more generic
declare -A motion_config_replace
motion_config_replace[replace_me_with_database_path]=$app_database
motion_config_replace[replace_me_with_log_path]=$app_service_log
motion_config_replace[replace_me_with_fusion_properties_path]=$fusion_properties
motion_config_replace[replace_me_with_user]=$app_user
motion_config_replace[replace_me_with_password]=$app_user_password
motion_config_replace[replace_me_with_smtp_server]=$smtp_server
motion_config_replace[replace_me_with_smtp_username]=$smtp_username
motion_config_replace[replace_me_with_smtp_password]=$smtp_password
motion_config_replace[replace_me_with_smtp_sender]=$smtp_sender
motion_config_file=$(create_file_from_template "$app_config_template" "motion_config_replace")
mv $motion_config_file $app_service_dir
# Create symlink for service config at top of motion_alerter directory
ln -s $app_service_config $app_dir/config.ini

# Create and install the configuration file for the civetweb webserver
declare -A civetweb_config_replace
civetweb_config_replace[replace_me_with_port]=$civetweb_port
civetweb_config_replace[replace_me_with_root_dir]=$civetweb_root_dir
civetweb_config_replace[replace_me_with_service_config]=$app_service_config
civetweb_config_file=$(create_file_from_template "$civetweb_config_template" "civetweb_config_replace")
mv $civetweb_config_file $civetweb_service_dir

# Create the service for civetweb
echo
echo "Installing civetweb_$app_name service."
declare -A civetweb_service_replace
civetweb_service_replace[replace_me_with_working_directory]=$civetweb_service_dir
civetweb_service_replace[replace_me_with_executable]="$civetweb_service_dir/civetweb $civetweb_service_dir/civetweb.conf"
civetweb_service_file=$(create_file_from_template "$civetweb_service_template" "civetweb_service_replace")
install_service $civetweb_service_file
rm $civetweb_service_file

# Restart fusion to load the app
echo
echo "Restarting Fusion."
systemctl restart ${FUSION_PRODUCT}

# Create the service for the motion alert notifier
# Restart motion_alerter after Fusion restarts, otherwise existing installs
# with pre-configured alerts can get spammed error messages in the logs
# about Fusion being unavailable
echo
echo "Installing $app_name service."
declare -A motion_service_replace
motion_service_replace[replace_me_with_working_directory]=$app_service_dir
motion_service_replace[replace_me_with_executable]="$app_service_dir/$app_name.py"
motion_service_file=$(create_file_from_template "$app_service_template" "motion_service_replace")
install_service $motion_service_file
rm $motion_service_file

echo
echo "$app_name has been installed to $app_dir."
