app_name=motion_alerter
app_service=$app_name
civetweb_service=civetweb_$app_name

if [[ -z $FUSION_PRODUCT ]]; then
    FUSION_PRODUCT="fusion"
fi

function err_exit()
{
    echo "Failed: $1"
    exit 1
}

function check_sudo()
{
    # Check that we have root privledges
    if [[ $EUID -ne 0 ]]; then
        err_exit "needs to be run with root privledges."
    fi
}

function check_deps_and_install()
{
    if which apt-get &> /dev/null; then
        pkg_mgr='apt-get'
    elif which yum &> /dev/null; then
        pkg_mgr='yum'
        # On RHEL/Centos 7, jq is not in the main repos, so add epel if needed
        version_id=$(egrep ^VERSION_ID= /etc/os-release | cut -d'"' -f2)
        if [[ $version_id -eq 7 ]] && ! rpm -q epel-release --quiet; then
            sudo yum -y install epel-release --quiet
        fi
    else
        err_exit "unable to determine package manager"
    fi

    for dep in $@; do
        which $dep &> /dev/null || eval "$pkg_mgr -y install $dep &> /dev/null"
        if [[ $? -ne 0 ]]; then
            err_exit "failed to install '$dep' package, must install manually to continue."
        fi
    done
}

function install_service()
{
    local service_file="$1"
    local service_name="$(basename $service_file)"

    cp "$service_file" "/etc/systemd/system/$service_name"

    systemctl enable $service_name
    systemctl daemon-reload
    systemctl restart $service_name
}

function create_file_from_template()
{
    local template_file="$1"
    local replace_map_name="$2"

    # Create a temp file from the template file that we will modify
    local file_name=$(basename -s .template "$template_file")
    local tmp_file=/tmp/$file_name
    cp "$template_file" "$tmp_file"

    # Create an associative from name of array passed in
    local replace_map_declare=$(declare -p "$replace_map_name")
    eval "local -A replace_map="${replace_map_declare#*=}

    # replace all the strings in the tmp file
    for replace_str in "${!replace_map[@]}"; do
        sed -i "s%$replace_str%${replace_map[$replace_str]}%" $tmp_file
    done
    
    # Return the path to the created file
    echo -n "$tmp_file"
}

function get_property_value()
{
    local property=$1

    local fusion_properties=/etc/opt/${FUSION_PRODUCT}/${FUSION_PRODUCT}.properties
    local property_line=$(grep "^\s*$property" "$fusion_properties" | tail -1)
    local property_value=$(echo $property_line | sed -E "s/^\s*$property\s*[:=]\s*(.*$)/\1/")

    # Return the property value
    echo -n "$property_value"
}

function get_fusion_apps_dir()
{
    local path=$(get_property_value "fusion.apps.path")

    pushd /etc/opt/${FUSION_PRODUCT} &> /dev/null
    # Resolve the path to absolute if it's relative.
    # If the path does not exist, readlink returns nothing
    local abs_path=$(readlink -e $path 2> /dev/null)
    popd &> /dev/null

    echo -n "$abs_path"
}

function get_fusion_url()
{
    local fusion_user="$1"
    local fusion_pass="$2"
    local port=$(get_property_value "listening.port")
    local protocol=$(get_property_value "listening.protocol")

    echo -n "${protocol}://${fusion_user}:${fusion_pass}@localhost:${port}"
}

function get_fusion_user_id()
{
    local fusion_user="$1"
    local fusion_pass="$2"
    local user_name="$3"

    local user_id=$(curl -ks "$(get_fusion_url $fusion_user $fusion_pass)/service/users" | \
        jq '.[] | select (.username == "$user_name") | .id' | tr -d '"')

    echo -n "$user_id"
}

function create_fusion_user()
{
    local fusion_user="$1"
    local fusion_pass="$2"
    local user="$3"
    local pass="$4"

    local fusion_endpoint="$(get_fusion_url $fusion_user $fusion_pass)/service"

    # Get admin group from fusion
    local admin_group_id=$(curl -ks "${fusion_endpoint}/permissions" | \
        jq '.[] | select (.name == "Administrator") | .id' | tr -d '"')

    # Determine if the user already exists in the Administrator group
    #  and create the user if they are not in the Adminitstrator group
    local user_from_fusion=$(curl -ks "${fusion_endpoint}/permissions/$admin_group_id" | \
        jq ".users[] | select(.username == \"$user\") | .username" | tr -d '"')
    if [[ "$user_from_fusion" != "$user" ]]; then

        # Create the user
        return_code=$(curl -ksf -o /dev/null -w "%{http_code}" \
            "${fusion_endpoint}/users" \
            -H 'Content-Type: application/json;charset=UTF-8' \
            --data-binary "{\"groups\":[{\"id\":\"$admin_group_id\"}], \
                \"password\":\"$pass\",\"username\":\"$user\"}" \
            --compressed)

        if [[ $return_code -ge 400 ]]; then
            err_exit "Could not create the user: $user"
        fi
    fi
}

function remove_fusion_user()
{
    local fusion_user=$1
    local fusion_pass=$2
    local user=$3

    local fusion_endpoint="$(get_fusion_url $fusion_user $fusion_pass)/service"

    # Get admin group from fusion
    local user_id=$(curl -ks "${fusion_endpoint}/users" | \
        jq ".[] | select (.username == \"$user\") | .id" | tr -d '"')

    if [[ ! -z "$user_id" ]]; then
        # Delete the user
        return_code=$(curl -X "DELETE" -ksf -o /dev/null -w "%{http_code}" \
            "${fusion_endpoint}/users/${user_id}")

        if [[ $return_code -ge 400 ]]; then
            err_exit "Could not delete the user: $user"
        fi
    fi
}
