import sqlite3
import json
import os
import sys

from abc import ABC, abstractmethod
from config import DB_PATH

class Route(ABC):
    """Abstract base class for a RESTful CGI application.

    All subclasses of Route shall return a (response body, status code) tuple.
    """

    def __init__(self):
        self.db = sqlite3.connect(DB_PATH, timeout=15)
        self.db.execute('PRAGMA foreign_keys = ON')
        self.db.execute('PRAGMA journal_mode = WAL')

    def __del__(self):
        self.db.close()

    @staticmethod
    def get_content(as_json=True):
        """Obtain CGI request content.

        If 'as_json' is True, return content as a JSON object,
        otherwise return the plain content.
        """
        content = sys.stdin.read(int(os.environ['CONTENT_LENGTH']))
        if as_json:
            return json.loads(content)
        return content

    @abstractmethod
    def get(self, resource):
        pass

    @abstractmethod
    def post(self):
        pass

    @abstractmethod
    def put(self, resource):
        pass

    @abstractmethod
    def delete(self, resource):
        pass

class RecipientGroup(Route):
    def __init__(self):
        super().__init__()

    def get(self, resource=None):
        if resource:
            return self._get_one(resource)
        return self._get_all()

    def post(self):
        """Bulk post all recipient groups.

        Any entries in the database that are not in the request are
        removed.
        """
        recipient_groups = self.get_content()
        with self.db:
            recipient_group_names = []
            for recipient_group in recipient_groups:
                rg_id = recipient_group['id']
                name = recipient_group['name']
                if not name:
                    raise ValueError('Cannot set empty recipient group names')
                recipients = RecipientGroup._recipients_to_string(recipient_group['recipients'])
                if rg_id:
                    self.db.execute('''
                    UPDATE recipient_group
                       SET name=?, recipients=?
                     WHERE id=?
                     ''', (name, recipients, int(rg_id)))
                else:
                    self.db.execute('''
                    INSERT INTO recipient_group(name, recipients)
                    VALUES(?,?)
                    ''', (name, recipients))
                recipient_group_names.append(name)
            self.db.execute('''
            DELETE FROM recipient_group WHERE name NOT IN ({})
             '''.format(','.join(['?'] * len(recipient_group_names))),
             recipient_group_names)
        return (self._get_all()[0], 201)

    def put(self, resource):
        """Update the specified recipient group.
        """
        recipient_group = self.get_content()
        with self.db:
            name = recipient_group['name']
            if not name:
                raise ValueError('Cannot set empty recipient group names')
            recipients = RecipientGroup._recipients_to_string(recipient_group['recipients'])
            self.db.execute('''
            UPDATE recipient_group
               SET name=?, recipients=?
             WHERE id=?
             ''', (name, recipients, resource))
        return self._get_one(resource)

    def delete(self, resource):
        """Delete the specified recipient group.
        """
        with self.db:
            self.db.execute('''
            DELETE FROM recipient_group WHERE id=?
             ''', (resource,))
        return ({}, 204)

    @staticmethod
    def _recipients_to_string(recipients):
        """Internal: validate recipients and convert to string.

        If 'recipients' is empty return None, otherwise return a comma separated
        string of recipients.

        If any values in 'recipients' contain an empty string, those values are silently
        removed.
        """
        # remove any empty strings from recipients list
        recipients = list(filter(None, recipients))
        if not recipients:
            return None
        return ','.join(recipients)

    def _get_all(self):
        """Internal: get all recipient groups.
        """
        return (json.dumps([
                {
                    'id': int(rg_id),
                    'name': name,
                    'recipients': recipients.split(',') if recipients else []
                } for rg_id, name, recipients in self.db.execute(
                    '''SELECT * FROM recipient_group''')
            ]), 200)

    def _get_one(self, resource):
        """Internal: get the specified recipient group.
        """
        result = self.db.execute('''
        SELECT name, recipients
          FROM recipient_group
         WHERE id=?
        ''', (resource,)).fetchone()
        if not result:
            raise ValueError('Recipient group ID={} does not exist'
                             .format(resource))
        name, recipients = result
        recipient_group = {
            'id': resource,
            'name': name,
            'recipients': recipients.split(',') if recipients else []
        }
        return (json.dumps(recipient_group), 200)

class Schedule(Route):
    def __init__(self):
        super().__init__()

    def get(self, resource=None):
        if resource:
            return self._get_one(resource)
        return self._get_all()

    def post(self):
        """Bulk post all schedules.

        Any entries in the database that are not in the request are
        removed.
        """
        schedules = self.get_content()
        with self.db:
            schedule_names = []
            for schedule in schedules:
                Schedule._assert_time(schedule['start'],
                                      schedule['stop'])
                s_id = schedule['id']
                name = schedule['name']
                if not name:
                    raise ValueError('Cannot set empty schedule names')
                start = schedule['start']
                stop = schedule['stop']
                days = Schedule._days_to_string(schedule['days'])
                if s_id:
                    self.db.execute('''
                    UPDATE schedule
                       SET name=?, start=?, stop=?, days=?
                     WHERE id=?
                     ''', (name, start, stop, days, int(s_id)))
                else:
                    self.db.execute('''
                    INSERT INTO schedule(name, start, stop, days)
                    VALUES(?,?,?,?)
                    ''', (name, start, stop, days))
                schedule_names.append(name)
            self.db.execute('''
            DELETE FROM schedule WHERE name NOT IN ({})
             '''.format(','.join(['?'] * len(schedule_names))),
             schedule_names)
        return (self._get_all()[0], 201)

    def put(self, resource):
        """Update the specified schedule.
        """
        schedule = self.get_content()
        Schedule._assert_time(schedule['start'],
                              schedule['stop'])
        with self.db:
            name = schedule['name']
            if not name:
                raise ValueError('Cannot set empty schedule names')
            start = schedule['start']
            stop = schedule['stop']
            days = Schedule._days_to_string(schedule['days'])
            self.db.execute('''
            UPDATE schedule
               SET name=?, start=?, stop=?, days=?
             WHERE id=?
             ''', (name, start, stop, days, resource))
        return self._get_one(resource)

    def delete(self, resource):
        """Delete the specified schedule.
        """
        with self.db:
            self.db.execute('''
            DELETE FROM schedule WHERE id=?
             ''', (resource,))
        return ({}, 204)
    
    @staticmethod
    def _days_to_string(days):
        """Internal: validate days and convert to string.

        If 'days' is empty return None, otherwise return a comma separated
        string of days.

        If any values in 'days' contain an empty string, those values are silently
        removed.
        """
        # remove any empty strings from days list
        days = list(filter(None, days))
        if not days:
            return None
        day_set = {
            'monday',
            'tuesday',
            'wednesday',
            'thursday',
            'friday',
            'saturday',
            'sunday'
        }
        valid_days = []
        # validate days list, remove any duplicates first
        for day in list(dict.fromkeys(days)):
            day = day.lower()
            if day not in day_set:
                raise ValueError('{} uses an incorrect format'.format(day))
            valid_days.append(day)
        return ','.join(valid_days)

    @staticmethod
    def _assert_time(*time_vals):
        """Internal: validate start/stop times.
        """
        if not time_vals or len(time_vals) != 2:
            raise ValueError('Invalid start/stop')
        for time in time_vals:
            if time < 0 or time > 1439:
                raise ValueError('Start/Stop valid range:'
                                 '[0,1439] (in minutes)')

    def _get_all(self):
        """Internal: get all schedules.
        """
        return (json.dumps([
                {
                    'id': int(s_id),
                    'name': name,
                    'start': start,
                    'stop': stop,
                    'days': days.split(',') if days else []
                } for s_id, name, start, stop, days in self.db.execute('''
                    SELECT * FROM schedule''')
            ]), 200)

    def _get_one(self, resource):
        """Internal: get the specified schedule.
        """
        result = self.db.execute('''
        SELECT name, start, stop, days
          FROM schedule
         WHERE id=?
         ''', (resource,)).fetchone()
        if not result:
            raise ValueError('Schedule ID={} does not exist'
                             .format(resource))
        name, start, stop, days = result
        schedule = {
            'id': resource,
            'name': name,
            'start': int(start),
            'stop': int(stop),
            'days': days.split(',') if days else []
        }
        return (json.dumps(schedule), 200)

class AlertConfiguration(Route):
    def __init__(self):
        super().__init__()

    def get(self, resource):
        if resource:
            return self._get_one(resource)
        return self._get_all()

    def post(self):
        """Bulk post all alert configurations.

        Any entries in the database that are not in the request are
        removed.
        """
        alert_configs = self.get_content()
        with self.db:
            orchid_uuids = []
            existing_uuids = set(
                orchid_uuid for orchid_uuid, in self.db.execute(
                    '''SELECT uuid FROM orchid'''))
            for alert_config in alert_configs:
                orchid_uuid = alert_config['orchidId']
                if not orchid_uuid:
                    raise ValueError('Cannot set an empty Orchid UUID')
                recipient_group_id = alert_config['recipientGroupId']
                if orchid_uuid in existing_uuids:
                    self.db.execute('''
                    UPDATE orchid SET recipient_group_id=? WHERE uuid=?
                     ''', (recipient_group_id, orchid_uuid))
                else:
                    self.db.execute('''
                    INSERT INTO orchid(uuid, recipient_group_id)
                    VALUES(?,?)
                    ''', (orchid_uuid, recipient_group_id))
                self._update_child_tables(alert_config)
                orchid_uuids.append(alert_config['orchidId'])
            diff = existing_uuids.difference(orchid_uuids)
            if diff:
                self.db.execute('''
                DELETE FROM orchid WHERE uuid IN ({})
                 '''.format(','.join(['?'] * len(diff))),
                 diff)
        return (self._get_all()[0], 201)
    
    def put(self, resource):
        """Update the specified alert configurtion.
        """
        alert_config = self.get_content()
        with self.db:
            recipient_group_id = alert_config['recipientGroupId']
            self.db.execute('''
            UPDATE orchid SET recipient_group_id=? WHERE uuid=?
             ''', (recipient_group_id, resource))
            self._update_child_tables(alert_config,
                    resource=resource)
        return self._get_one(resource)

    def delete(self, resource):
        """Delete the specified alert configuration.
        """
        with self.db:
            self.db.execute('''
            DELETE FROM orchid WHERE uuid=?
             ''', (resource,))
        return ({}, 204)

    def _get_all(self):
        """Internal: get all alert configurations.
        """
        return (json.dumps([
                self._load_alert_config(
                    orchid_uuid=orchid_uuid,
                    recipient_group_id=recipient_group_id
                ) for orchid_uuid, recipient_group_id in self.db.execute('''
                    SELECT * FROM orchid''')
            ]), 200)

    def _get_one(self, resource):
        """Internal: get the specified alert configuration.
        """
        result = self.db.execute('''
        SELECT recipient_group_id FROM orchid WHERE uuid=?
         ''', (resource,)).fetchone()
        if not result:
            raise ValueError('Orchid UUID={} does not exist'
                             .format(resource))
        recipient_group_id, = result
        return (json.dumps(
            self._load_alert_config(
                orchid_uuid=resource,
                recipient_group_id=recipient_group_id)
            ), 200)

    def _load_alert_config(self, orchid_uuid, recipient_group_id):
        """Internal: build the alert configuration resource.
        """
        recipient_group_id = int(recipient_group_id) if recipient_group_id else None
        return {
            'orchidId': orchid_uuid,
            'recipientGroupId': recipient_group_id,
            'cameraIds': [
                camera_id for camera_id, in self.db.execute('''
                SELECT camera_id FROM orchid_camera WHERE orchid_uuid=?
                 ''', (orchid_uuid,))
            ],
            'scheduleIds': [
                schedule_id for schedule_id, in self.db.execute('''
                SELECT schedule_id FROM orchid_schedule WHERE orchid_uuid=?
                 ''', (orchid_uuid,))
            ]
        }

    def _update_child_tables(self, alert_config, resource=None):
        """Internal: update all children of the parent 'orchid' table.

        Note that this function runs within a transaction.
        """
        orchid_uuid = resource if resource else alert_config['orchidId']
        self._clean_insert({
            'orchid_uuid': orchid_uuid,
            'inserts': [
                {
                    'values': alert_config['cameraIds'],
                    'table': 'orchid_camera',
                    'column': 'camera_id'
                },
                {
                    'values': alert_config['scheduleIds'],
                    'table': 'orchid_schedule',
                    'column': 'schedule_id'
                }
            ]
        })

    def _clean_insert(self, db_update):
        """Internal

        Remove any rows in the database that are not in the request, then
        insert all entries from request.

        Note that this function runs within a transaction.
        """
        for insert in db_update['inserts']:
            # get what's in the database
            in_db = set(col for col, in self.db.execute('''
            SELECT {column} FROM {table} WHERE orchid_uuid=?
            '''.format(column=insert['column'], table=insert['table']),
            (db_update['orchid_uuid'],)))
            # find what's in the database that's not in the update
            diff = in_db.difference(insert['values'])
            if diff: # if something changed, remove the stale rows
                self.db.execute('''
                DELETE FROM {table}
                 WHERE orchid_uuid=?
                   AND {column} IN ({targets})
                   '''.format(
                       table=insert['table'],
                       column=insert['column'],
                       targets=','.join(['?'] * len(diff))),
                   (db_update['orchid_uuid'],) + tuple(diff))
            self.db.executemany('''
            INSERT OR IGNORE INTO {table}(orchid_uuid, {column})
            VALUES(?,?)
            '''.format(table=insert['table'], column=insert['column']),
            ( (db_update['orchid_uuid'], val) for val in insert['values'] ))
